#
# What is this?
#   This is a fabric file that is a simple test runner for cumulus
#   integration tests for NCLU. It will run tests against branches
#   continuously and record results of those runs.
#
# Why?
#   TRDB only tests dev and some tests only run once a week. This
#   causes a very bad developer experience. Test failures should
#   be identified as early in the process as possible.
#   Additionally, running tests manually can be cumbersome since
#   you need to remember all the commands and to update all the
#   test packages. This prevents mistakes and gives you a way
#   to continuously test NCLU.
#
# What does it do?
#
#   - run tests against tags and nclu branches
#   - update hard node
#   - build nclu deb
#   - install nclu deb
#   - test reporting
#   - coverage reporting
#
# How?
#   Copy this file into a directory and modify the variables at
#   the top of the file.
#
# Run tests:
#   I would recommend using something like screen to run the tests in:
#
#      fab run_tests
#
# Get results:
#
#      fab details
#
#      fab details:failures
#
#      fab details:failures,dev-next
#
#      fab show:nclu.nclu_vxlan_tests.json,dev-next
#      fab show_history:nclu.nclu_vxlan_tests.json,dev-next
#
# Others:
#   - build_branch
#   - git_update
#   - install_db
#   - summarize


from __future__ import with_statement  # noqa

import json
import os
import random
import re
import sys
import time
import traceback
from datetime import datetime
from fnmatch import fnmatch
from pprint import pformat

from fabric.api import cd, env, get, local, put, run, settings, sudo, lcd
from fabric.contrib import files
from tabulate import tabulate

# hard node to run tests again
# you should have used ssh-copy-id against this host
env.hosts = ['cel-redxp-10']
env.user = 'cumulus'
env.password = 'CumulusLinux!'

TAG = 'nclu-smoke'
LOG_LEVEL = 'WARNING'
IMAGES = '/home/build/ci/cumulus/CumulusLinux-3.0.0/latest-vx-amd64/build/images/'  # noqa
BRANCHES = ['dev-next', 'dev']

HOME = '/home/nathan'
NCLU_DIR = '/home/nathan/nclu'
REPO_DIR = '/home/nathan/repo'


TESTINFRA = os.environ['TESTINFRA']
CL_CAFE = os.path.join(TESTINFRA, 'cl-cafe')
CL_TESTS = os.path.join(TESTINFRA, 'cl-tests')
CL_TOOLS = os.path.join(TESTINFRA, 'cl-tools')

sys.path.extend([
    CL_CAFE, CL_TESTS, CL_TOOLS
])


CUR_DIR = os.path.dirname(os.path.realpath(__file__))
RES_DIR = os.path.join(CUR_DIR, 'results')
COV_DIR = os.path.join(CUR_DIR, 'cover')
SETTINGS_FILE = os.path.join(CUR_DIR, '.settings.json')
OPTIONS = {
    'prune': 20,  # how many test results to keep
    'success': 60 * 60 * 6,  # how often to run successful tests again
    'failed': 60 * 60 * 2,  # how often to re-run failed tests
    'active': True,
    'coverage': False
}

if not os.path.exists(RES_DIR):
    os.mkdir(RES_DIR)
if not os.path.exists(COV_DIR):
    os.mkdir(COV_DIR)


def _update_settings():
    if os.path.exists(SETTINGS_FILE):
        with open(SETTINGS_FILE) as fi:
            OPTIONS.update(json.loads(fi.read()))


_update_settings()


def _get_all_test_files():
    results = []
    tests_path = os.path.join(CL_TESTS, 'tests')
    for dir_name in os.listdir(tests_path):
        if dir_name in ('config', 'configs', 'examples',
                        'fuse', 'netq', 'lib', 'specs'):
            continue
        test_group_path = os.path.join(tests_path, dir_name)
        if not os.path.isdir(test_group_path):
            continue
        for filename in os.listdir(test_group_path):
            if (not filename.lower().endswith('.py') or
                    filename == '__init__.py'):
                continue
            test_path = os.path.join(test_group_path, filename)
            if os.path.isfile(test_path):
                results.append(test_path)
    # we should randomize order that we run these...
    random.shuffle(results)
    return results


def _get_test_result_filepath(filepath):
    tests_path = os.path.join(CL_TESTS, 'tests')
    test_name = filepath[len(tests_path) + 1:].replace(
        '.py', '').replace('/', '.')
    return os.path.join(RES_DIR, test_name) + '.json'


def _get_test_results(filepath):
    result_filepath = _get_test_result_filepath(filepath)
    if os.path.exists(result_filepath):
        with open(result_filepath) as fi:
            return json.loads(fi.read())
    return {}


def _save_test_results(filepath, data):
    result_filepath = _get_test_result_filepath(filepath)
    for key in [k for k in data.keys()]:
        data[key] = data[key][-(OPTIONS['prune']):]
    with open(result_filepath, 'w') as fi:
        return fi.write(json.dumps(data))


def git_update():
    with settings():
        for filepath in (CL_CAFE, CL_TESTS, CL_TOOLS):
            local('cd ' + filepath + ' && git pull')


def run_test(test_file, branch, full=False):
    data = _get_test_results(test_file)
    if len(data.get(branch, [])) > 0 and not False:
        # check last success, see if it is time to run again
        last = data[branch][-1]
        seconds = int(time.time() - last['timestamp'])
        if last['return_code'] != 0:
            if seconds < OPTIONS['failed']:
                print('Skipping failed test: {} - last failed: {}'.format(
                    test_file, _time_since(last['timestamp'])))
                return
        else:
            if seconds < OPTIONS['success']:
                print('Skipping test: {} - last success: {}'.format(
                    test_file, _time_since(last['timestamp'])))
                return

    start = time.time()
    result = local(' '.join([
        'runtests.sh', '-d', env.hosts[0], '--vx-img',
        '-l', LOG_LEVEL,
        '-t', TAG,
        '-i', IMAGES,
        '--repo', REPO_DIR,
        test_file
    ]), capture=True)
    end = time.time()
    if result.return_code != 0:
        print('Failed {} with output \nstdout:\n{}\nstderr:\n{}'.format(
            result.command, result.stdout, result.stderr))
    if branch not in data:
        data[branch] = []
    data[branch].append({
        'timestamp': time.time(),
        'command': result.command,
        'return_code': result.return_code,
        'stdout': result.stdout,
        'stderr': result.stderr,
        'duration': int(end - start),
        'tag': TAG
    })
    _save_test_results(test_file, data)


def run_tests(full=False, match=None, branch=None, continuous=None):
    while True:
        try:
            _run_tests(full, match, branch, continuous)
        except Exception:
            traceback.print_exc()
            print('Unhandled exception, take a break and try again')
            time.sleep(60 * 1)


def _run_tests(full, match, branch, continuous):
    for current_branch in BRANCHES:
        if branch is not None and branch != current_branch:
            continue
        store_coverage()
        build_branch(current_branch)
        with settings(warn_only=True):
            git_update()
            sudo('apt-get update && apt-get upgrade -y')
            install_deb()
            test_files = _get_all_test_files()
            for test_file in test_files:
                if match is not None and not fnmatch(
                        test_file, '*' + match + '*'):
                    continue
                _update_settings()
                if not OPTIONS['active']:
                    print('Running tests disabled...')
                    restore_netd()
                    return
                run_test(test_file, current_branch, full)
            time.sleep(1)
    if continuous is not None and continuous.lower() == 'false':
        return


def disable_tests():
    OPTIONS['active'] = False
    with open(SETTINGS_FILE, 'w') as fi:
        fi.write(json.dumps(OPTIONS))


def enable_tests():
    OPTIONS['active'] = True
    with open(SETTINGS_FILE, 'w') as fi:
        fi.write(json.dumps(OPTIONS))


def kill_netd():
    # see if we have existing coverage report running.
    # kill if it is, them move report data
    found = True
    while found:
        found = False
        for line in run('ps auxx | grep netd').splitlines():
            if 'coverage run' in line:
                found = True
                # kill it!
                pid = line.split()[1]
                sudo('kill {}'.format(pid))
                time.sleep(1)
                break


def restore_netd():
    sudo('systemctl stop netd')
    kill_netd()
    if OPTIONS['coverage']:
        with cd('/home/cumulus'):
            sudo('pip install coverage')
            sudo('touch /.coverage_enabled')
            sudo(' '.join([
                 'nohup /usr/local/bin/coverage run ',
                 '--include=*nclu* /usr/sbin/netd --debug ',
                 '&> /dev/null &']))
    else:
        sudo('rm /.coverage_enabled*')
        sudo('systemctl reset-failed netd')
        sudo('systemctl start netd')


def store_coverage():
    with lcd(NCLU_DIR):
        result = local('git rev-parse --abbrev-ref HEAD --', capture=True)
    running_branch = result.split()[0]

    with settings(warn_only=True):
        run('net _stop_coverage_', warn_only=True)
        if files.exists('/home/cumulus/.coverage'):

            sudo('pip install coverage')
            with cd('/home/cumulus'):
                run('coverage report -m > report')
                run('coverage html')

            coverage_filepath = os.path.join(
                COV_DIR, '{}-{}'.format(
                    running_branch, datetime.utcnow().isoformat()))
            local('mkdir {}'.format(coverage_filepath))

            get('/home/cumulus/.coverage', '{}/.coverage'.format(
                coverage_filepath))
            get('/home/cumulus/report', '{}/report'.format(coverage_filepath))
            run('tar -zcvf htmlcov.tar.gz htmlcov')
            get('/home/cumulus/htmlcov.tar.gz',
                '{}/htmlcov.tar.gz'.format(coverage_filepath))

            sudo('rm -rf /home/cumulus/.coverage*')
            sudo('rm -rf /home/cumulus/htmlcov*')
            sudo('rm -rf /home/cumulus/report*')


def build_branch(name):
    local('rm -rf {}/nclu_*'.format(HOME))
    local('rm -rf {}/debian/nclu_*'.format(REPO_DIR))
    with lcd(NCLU_DIR):
        local('git pull')
        local('git checkout ' + name)

    with settings(host_string='nathan@marvel-07'):
        with cd(NCLU_DIR):
            run('sbuild -j 4 -A '
                '-d `cl-schroot.sh --official`-dev-full-amd64')
    local('mv {}/nclu_*.deb {}/debian/'.format(HOME, REPO_DIR))


def install_deb():
    sudo('rm -rf /home/cumulus/nclu*')
    put(local_path='{}/debian/*.deb'.format(REPO_DIR),
        remote_path='/home/cumulus/')
    sudo('apt-get remove --purge nclu -y')
    sudo('rm -rf /usr/local/lib/python2.7/dist-packages/nclu/*')
    sudo('dpkg -i /home/cumulus/*.deb')
    restore_netd()


def _get_test_details(filename):
    filepath = os.path.join(RES_DIR, filename)
    with open(filepath) as fi:
        data = json.loads(fi.read())
        result = {}
        for key, value in data.items():
            if key not in result:
                result[key] = []
            for test_result in value:
                match = re.search(
                    'Ran (?P<count>[\d]*) tests? in [\d\.]*s',
                    test_result['stderr'])
                if not match:
                    continue
                num_tested = int(match.groupdict()['count'])
                errors = failures = 0
                match = re.search(
                    'failures\=(?P<count>[\d]*)',
                    test_result['stderr'])
                if match:
                    failures = int(match.groupdict()['count'])
                match = re.search(
                    'errors\=(?P<count>[\d]*)',
                    test_result['stderr'])
                if match:
                    errors = int(match.groupdict()['count'])
                test_result.update({
                    'success': max(num_tested - errors - failures, 0),
                    'failures': failures,
                    'errors': errors
                })
                result[key].append(test_result)
        return result


def _get_details():
    file_data = {}
    for filename in os.listdir(RES_DIR):
        file_data[filename] = _get_test_details(filename)
    return file_data


def summarize():
    results = {}
    for filename, file_data in _get_details().items():
        for branch, data in file_data.items():
            if branch not in results:
                results[branch] = {
                    'success': 0,
                    'failures': 0,
                    'errors': 0,
                    'files': 0
                }
            if len(data) > 0:
                results[branch]['files'] += 1
                results[branch]['success'] += data[-1]['success']
                results[branch]['failures'] += data[-1]['failures']
                results[branch]['errors'] += data[-1]['errors']

    print(pformat(results))


def _plural(val, word):
    if val > 1:
        return word + 's'
    return word


def _time_since(timestamp):
    diff = int(time.time() - timestamp)
    if diff < 60:
        return '{} {}'.format(diff, _plural(diff, 'sec'))
    if diff < (60 * 60):
        val = diff / 60
        return '{} {}'.format(val, _plural(val, 'min'))
    if diff < (60 * 60 * 24):
        val = diff / 60 / 60
        return '{} {}'.format(val, _plural(val, 'hr'))
    val = diff / 60 / 60 / 24
    return '{} {}'.format(val, _plural(val, 'day'))


def details(filtered=None, branch=None):
    output = []
    details = _get_details()
    for filename in sorted(details.keys()):
        file_data = details[filename]
        for the_branch, data in file_data.items():
            if len(data) == 0:
                continue
            if branch is not None and the_branch != branch:
                continue
            last_run = data[-1]
            last_failure = last_success = None
            for item in reversed(data):
                if item['failures'] == 0 and item['errors'] == 0:
                    if last_success is None:
                        last_success = _time_since(item['timestamp'])
                else:
                    if last_failure is None:
                        last_failure = _time_since(item['timestamp'])
            if (filtered == 'failures' and last_run['errors'] == 0 and
                    last_run['failures'] == 0):
                continue
            if (last_run['failures'] == 0 and
                    last_run['errors'] == 0
                    and last_run['success'] == 0):
                # no tests, ignore
                continue
            output.append([
                '{},{}:'.format(filename, the_branch),
                _time_since(last_run['timestamp']),
                last_run['success'],
                _color_if(last_run['failures'], _red),
                _color_if(last_run['errors'], _red),
                last_failure, last_success])
    print(tabulate(output, headers=[
        'Test', 'Last run', 'Success', 'Failed',
        'Errors', 'Last failure', 'Last success'
    ]))


def _get_test_failures(output):
    found = []
    capturing = False
    for line in output.splitlines():
        if (line.startswith('Ran ') or line.startswith('FAILED ') or
                'failures=' in line or 'errors=' in line):
            found.append(line)
            continue
        if line.startswith('FAIL:') or line.startswith('ERROR:'):
            capturing = True
        if capturing:
            if line.startswith('-------------------- >>'):
                capturing = False
            else:
                found.append(line)
    return '\n'.join(found)


def list_failures(test=None, branch=None):
    details = _get_details()
    for filename in sorted(details.keys()):
        if test is not None and test != filename:
            continue
        file_data = details[filename]
        for branch, data in file_data.items():
            if branch is not None and branch != branch:
                continue
            if len(data) == 0:
                continue
            last_run = data[-1]
            if last_run['failures'] == 0 and last_run['errors'] == 0:
                continue
            print('\n\n\nFailed: {}:{}\n{}'.format(
                filename, branch, _get_test_failures(last_run['stderr'])))


def show(test, branch='dev', item=0):
    filepath = os.path.join(RES_DIR, test)
    with open(filepath) as fi:
        data = json.loads(fi.read())
    item = data[branch][-(int(item) + 1)]
    print(item['stderr'])


def show_history(test, branch='dev'):
    data = _get_test_details(test)
    output = []
    for item in data[branch]:
        output.append([
            _time_since(item['timestamp']),
            item['success'],
            item['failures'],
            item['errors']
        ])
    print(tabulate(output, headers=[
        'When', 'Success', 'Failed', 'Errors'
    ]))


def _color_if(val, color, condition=None):
    if condition is not None:
        if condition:
            return color(val)
        return val
    if val > 0:
        return color(val)
    return val


def _red(val):
    return "\x1b[31m{}\x1b[39m".format(val)
