from nclu.tests.base import BaseTest
from nclu.tests.utils import read_data_file


class TestIfupdown2(BaseTest):

    def test_write_interface_does_not_produce_diff(self):
        # patches need to happen before modules are loaded
        import nclu.NetDaemon  # noqa
        from nclu.wrappers import ifupdown2
        from nclu import make_pending_diff

        for name in ('simple', 'simple2'):
            self.patcher.expect_file(
                ifupdown2.ETC_NETWORK_INTERFACES,
                read_data_file('etc/network/interfaces/{}'.format(name)))
            wrapper = ifupdown2.ConfigWrapper(False, False)
            self.assertTrue(not wrapper._save_or_show_pending(
                False, False)[1])
            diff = make_pending_diff(
                ifupdown2.ETC_NETWORK_INTERFACES,
                ifupdown2.INTERFACES_TMP, False)
            self.assertEquals(diff, '')
            # also make sure with persist but no
            # diff we do not write/reload
            self.assertTrue(not wrapper._save_or_show_pending(
                persist=False, verbose=False, has_diff=False)[1])

    def test_writes_correct_bridge_command(self):
        # patches need to happen before modules are loaded
        import nclu.NetDaemon  # noqa
        from nclu.wrappers import ifupdown2

        ifupdown2.build_doc_string([])
        self.patcher.expect_file(
            ifupdown2.ETC_NETWORK_INTERFACES,
            read_data_file('etc/network/interfaces/bridge'))
        self.patcher.expect_command(
            '/sbin/ifquery -a --raw',
            read_data_file('commands/ifquery/all-raw'))
        wrapper = ifupdown2.ConfigWrapper(False, False)
        commands, unsupported = wrapper.show_config_commands(True, None)
        assert 'net add interface swp5 bridge access 10' in commands
        assert len(unsupported) == 0

    def test_del_all_removes_peerlink(self):
        # patches need to happen before modules are loaded
        import nclu.NetDaemon  # noqa
        from nclu.wrappers import ifupdown2
        from nclu import make_pending_diff

        self.patcher.expect_file(
            ifupdown2.ETC_NETWORK_INTERFACES,
            read_data_file('etc/network/interfaces/peerlink'))
        self.patcher.expect_command(
            '/sbin/ifquery -a --raw',
            read_data_file('etc/network/interfaces/peerlink'))
        wrapper = ifupdown2.ConfigWrapper(False, False)
        wrapper.del_all(None)
        wrapper._save_or_show_pending(persist=False, verbose=False)
        diff = make_pending_diff(
                ifupdown2.ETC_NETWORK_INTERFACES,
                ifupdown2.INTERFACES_TMP, False)
        self.assertIn('-iface peerlink.4094', diff)

    def test_add_bond_tab_complete_works_with_mode_new_bond(self):
        import nclu.NetDaemon  # noqa
        from nclu.tab_completion import tab_complete_policer
        from nclu.wrappers import ifupdown2

        self.patcher.expect_file('/etc/netd.conf', '''
enable_ifupdown2=True
color_diffs=False''')
        self.patcher.expect_command(
            '/sbin/ifquery -a --raw',
            read_data_file('etc/network/interfaces/simple'))

        ifupdown2.initialize_global_ifupdown2(False, False)

        words = tab_complete_policer(
            {'<interface>': 'bond77', 'net': 'net',
             'add': 'add', 'bond': 'bond'},
            ['downdelay', 'lacp-bypass-allow', 'lacp-rate', 'miimon',
             'min-links', 'mode', 'slaves', 'updelay'],
            True, 'bond')
        self.assertIn('mode', words)
        self.assertIn('slaves', words)
        # only allow mode and slaves when bond not already defined
        self.assertEqual(len(words), 2)

    def test_add_bond_tab_complete_works_with_mode_new_bond_with_incomplete_tabbing(self):  # noqa
        import nclu.NetDaemon  # noqa
        from nclu.tab_completion import tab_complete_policer
        from nclu.wrappers import ifupdown2

        self.patcher.expect_file('/etc/netd.conf', '''
enable_ifupdown2=True
color_diffs=False''')
        self.patcher.expect_command(
            '/sbin/ifquery -a --raw',
            read_data_file('etc/network/interfaces/simple'))

        ifupdown2.initialize_global_ifupdown2(False, False)

        words = tab_complete_policer(
            {'miimon': 'miimon', '<interface>': 'bond77',
             'add': 'add', 'mode': 'mode', 'net': 'net',
             'min-links': 'min-links', 'bond': 'bond'},
            ['miimon', 'min-links', 'mode'],
            False, 'm')
        self.assertEqual(words, ['mode'])

    def test_add_bond_tab_complete_works_with_mode_existing_bond(self):
        import nclu.NetDaemon  # noqa
        from nclu.tab_completion import tab_complete_policer
        from nclu.wrappers import ifupdown2

        self.patcher.expect_file('/etc/netd.conf', '''
enable_ifupdown2=True
color_diffs=False''')
        self.patcher.expect_command(
            '/sbin/ifquery -a --raw',
            '''
auto lo
iface lo inet loopback

auto eth0
iface eth0 inet dhcp

auto swp1
iface swp1

auto bond77
iface bond77
    bond-slaves swp1
''')

        ifupdown2.initialize_global_ifupdown2(False, False)

        words = tab_complete_policer(
            {'<interface>': 'bond77', 'net': 'net',
             'add': 'add', 'bond': 'bond'},
            ['downdelay', 'lacp-bypass-allow', 'lacp-rate', 'miimon',
             'min-links', 'mode', 'slaves', 'updelay'],
            True, 'bond')
        self.assertIn('mode', words)
        self.assertIn('slaves', words)
        self.assertIn('downdelay', words)
        # allow all config types if we already have bond defined
        self.assertEqual(len(words), 8)

    def test_add_bond_mode_before_slaves_allowed(self):
        '''
        Should be able to add mode before slaves.
        '''
        # convince machinery that ifupdown2 enabled
        self.patcher.expect_file('/etc/netd.conf', '''
enable_ifupdown2=True
color_diffs=False''')
        self.patcher.expect_file(
            '/usr/share/ifupdown2/addons/foobar', '')
        self.patcher.expect_command(
            '/sbin/ifquery -a --raw',
            read_data_file('etc/network/interfaces/simple'))

        from nclu.NetDaemon import NetDaemon
        daemon = NetDaemon()
        success, result = self.execute_command(
            daemon, 'net add bond bond77 bond mode balance-xor')
        self.assertTrue(success)
        self.assertEqual(result, '')
        self.assertIn(
            '/run/nclu/ifupdown2/bond77',
            self.patcher.file_expectations)
        self.assertIn(
            'iface bond77',
            self.patcher.file_expectations['/run/nclu/ifupdown2/bond77'])
        self.assertIn(
            'bond-mode balance-xor',
            self.patcher.file_expectations['/run/nclu/ifupdown2/bond77'])

        # now add the slaves
        self.patcher.expect_file('/sys/class/net/swp1', '')
        self.patcher.expect_file('/sys/class/net/swp2', '')
        self.patcher.expect_file('/sys/class/net/swp3', '')

        success, result = self.execute_command(
            daemon, 'net add bond bond77 bond slaves swp1-3')
        self.assertTrue(success)
        self.assertEqual(result, '')
        self.assertIn(
            'bond-slaves swp1 swp2 swp3',
            self.patcher.file_expectations['/run/nclu/ifupdown2/bond77'])

    def test_not_add_interface_for_existing_breakout(self):
        self.patcher.expect_file('/etc/cumulus/ports.conf', """\
1=10G
2=10G
3=10G
""")
        self.patcher.expect_file('/etc/netd.conf', """\
enable_ifupdown2=True
enable_ports=True
color_diffs=False
""")
        self.patcher.expect_file('/sys/class/net/swp1', 'foobar')
        self.patcher.expect_file('/sys/class/net/swp2', 'foobar')
        self.patcher.expect_file('/usr/share/ifupdown2/addons/foobar', '')
        self.patcher.expect_file('/etc/network/interfaces', read_data_file('etc/network/interfaces/simple'))
        self.patcher.expect_command('/sbin/ifquery -a --raw', read_data_file('etc/network/interfaces/simple'))

        from nclu.NetDaemon import NetDaemon
        daemon = NetDaemon()

        success, _ = self.execute_command(daemon, "net add interface swp2 breakout 4x")
        self.assertTrue(success)

        # Cannot "breakout 4x" the interface again.
        # TODO - This is a no-op.  The command should not fail.
        success, _ = self.execute_command(daemon, "net add interface swp2 breakout 4x")
        self.assertFalse(success)

        # Cannot add the root interface if it is broken-out (pending).
        success, _ = self.execute_command(daemon, "net add interface swp2")
        self.assertFalse(success)

        success, _ = self.execute_command(daemon, "net add interface swp1")
        self.assertTrue(success)

    def test_not_add_interface_for_already_breakout(self):
        self.patcher.expect_file('/etc/cumulus/ports.conf', """\
1=10G
2=10G
3=10G
""")
        self.patcher.expect_file('/etc/netd.conf', """\
enable_ifupdown2=True
enable_ports=True
color_diffs=False
""")
        self.patcher.expect_file('/sys/class/net/swp1', 'foobar')
        config = """\
auto lo
iface lo inet loopback
auto eth0
iface eth0 inet dhcp
auto swp1s0
 iface swp1s0

 auto swp1s1
 iface swp1s1

 auto swp1s2
 iface swp1s2

 auto swp1s3
 iface swp1s3
"""
        self.patcher.expect_file('/usr/share/ifupdown2/addons/foobar', '')
        self.patcher.expect_file('/etc/network/interfaces', config)
        self.patcher.expect_command('/sbin/ifquery -a --raw', config)

        from nclu.NetDaemon import NetDaemon
        daemon = NetDaemon()

        # Cannot add the root interface if it is broken-out (running).
        success, _ = self.execute_command(daemon, 'net add interface swp1')
        self.assertFalse(success)
