# -*- coding: utf-8 -*-
import json

from nclu.tests.base import BaseTest
from nclu.tests.utils import read_data_file


class TestShow(BaseTest):

    def test_show_system(self):
        from nclu import netshow2
        self.patcher.expect_command('/usr/bin/platform-detect',
                                    output='cel,redstone_xp')
        self.patcher.expect_file(
            '/etc/lsb-release',
            ''''DISTRIB_ID="Cumulus Linux"
DISTRIB_RELEASE=3.6.2
DISTRIB_DESCRIPTION="Cumulus Linux 3.6.2"''')
        result = netshow2.net_show_system()
        self.assertTrue(result.startswith('''Cel REDSTONE
Cumulus Linux 3.6.2
Build: Cumulus Linux 3.6.2'''))

    def test_show_system_details(self):
        from nclu import netshow2
        self.patcher.expect_command('/usr/bin/platform-detect',
                                    output='cel,redstone_xp')
        self.patcher.expect_file(
            '/etc/lsb-release',
            ''''DISTRIB_ID="Cumulus Linux"
DISTRIB_RELEASE=3.6.2
DISTRIB_DESCRIPTION="Cumulus Linux 3.6.2"''')
        self.patcher.expect_file(
            '/usr/share/cumulus-platform/common/platform-info/accton.json',
            read_data_file('platform-info/accton.json'))
        self.patcher.expect_file(
            '/usr/share/cumulus-platform/common/platform-info/celestica.json',
            read_data_file('platform-info/celestica.json'))
        self.patcher.expect_file('/proc/uptime', '22341.37 82048.25')
        result = netshow2.net_show_system()
        self.assertTrue(result.startswith('''Cel REDSTONE
Cumulus Linux 3.6.2
Build: Cumulus Linux 3.6.2'''))
        self.assertTrue('CPU: ' in result)
        self.assertTrue('Uptime: ' in result)

    def test_show_system_json(self):
        from nclu import netshow2
        self.patcher.expect_command('/usr/bin/platform-detect',
                                    output='cel,redstone_xp')
        self.patcher.expect_file(
            '/etc/lsb-release',
            ''''DISTRIB_ID="Cumulus Linux"
DISTRIB_RELEASE=3.6.2
DISTRIB_DESCRIPTION="Cumulus Linux 3.6.2"''')
        self.patcher.expect_file('/proc/uptime', '22341.37 82048.25')
        result = json.loads(netshow2.net_show_system(use_json=True))
        self.assertEquals(result['platform']['model'], 'REDSTONE')
        self.assertEquals(result['platform']['vendor'], 'Cel')
        self.assertTrue(result['platform']['detected'])
        self.assertEquals(result['os-version'], '3.6.2')
        self.assertEquals(result['build'], 'Cumulus Linux 3.6.2')
        self.assertEquals(result['uptime'], '6:12:21.370000')

    def test_net_show_interfaces_subset_unicode_bug(self):
        from nclu import netshow2
        self.patcher.expect_command(
            ['/usr/bin/vtysh', '-c', 'show interface swp1'],
            output='')
        self.patcher.expect_command(
            '/usr/cumulus/bin/cl-netstat --json',
            output=read_data_file('commands/cl-netstat/1'))
        self.patcher.expect_command(
            '/sbin/ip -6 neighbor show',
            output=read_data_file('commands/ip/1'))
        self.patcher.expect_command(
            '/sbin/ip -4 neighbor show',
            output=read_data_file('commands/ip/2'))
        self.patcher.expect_command(
            'lspci -nn',
            output=read_data_file('commands/lspci/1'))
        self.patcher.expect_command(
            '/sbin/bridge -j -c vlan show',
            output='{}')
        self.patcher.expect_file(
            '/sys/class/net/swp1/ifalias', 'interface—0')

        result = netshow2.net_show_interface_subset(
            {}, ['swp1'], {}, None)
        self.assertTrue(u'interface\u20140' in result)

    def test_net_show_configuration_with_interface(self):
        from nclu.NetDaemon import NetDaemon
        from nclu.plugins import frr

        self.patcher.expect_file('/etc/netd.conf', '''
enable_ifupdown2=True
enable_frr=True
color_diffs=False''')
        self.patcher.expect_file(
            '/usr/share/ifupdown2/addons/foobar', '')
        self.patcher.expect_command(
            '/sbin/ifquery -a --raw',
            read_data_file('etc/network/interfaces/simple2'))
        self.patcher.expect_file(frr.ETC_FRR_CONF, '''
service integrated-vtysh-config
''')
        self.patcher.expect_command(
            '/usr/bin/vtysh -m -f {}'.format(frr.ETC_FRR_CONF),
            '''frr defaults datacenter
service integrated-vtysh-config
password cn321
line vty
 exec-timeout 0 0

end
router bgp 40201
 neighbor 3800:1:4::21 remote-as internal

end
''')

        daemon = NetDaemon()
        daemon.user = 'netd'
        success, result = self.execute_command(
            daemon, 'net show configuration interface swp1 json')
        self.assertTrue(success)
        self.assertIn('INTERFACE:swp1', result)


class TestBondMode(BaseTest):

    def test_get_balance_rr(self):
        self.patcher.expect_file(
            '/sys/class/net/bond77/bonding/mode', 'balance-rr 0')
        from nclu import netshow2
        self.assertEqual(netshow2.get_bond_mode('bond77'), 'Balance-RR')

    def test_get_active_backup(self):
        self.patcher.expect_file(
            '/sys/class/net/bond77/bonding/mode', 'active-backup 1')
        from nclu import netshow2
        self.assertEqual(netshow2.get_bond_mode('bond77'), 'Active-Backup')

    def test_get_balance_xor(self):
        self.patcher.expect_file(
            '/sys/class/net/bond77/bonding/mode', 'balance-xor 2')
        from nclu import netshow2
        self.assertEqual(netshow2.get_bond_mode('bond77'), 'Balance-XOR')

    def test_get_broadcast(self):
        self.patcher.expect_file(
            '/sys/class/net/bond77/bonding/mode', 'broadcast 3')
        from nclu import netshow2
        self.assertEqual(netshow2.get_bond_mode('bond77'), 'Broadcast')

    def test_get_802(self):
        self.patcher.expect_file(
            '/sys/class/net/bond77/bonding/mode', '802.3ad 4')
        from nclu import netshow2
        self.assertEqual(netshow2.get_bond_mode('bond77'), '802.3ad')
