import __builtin__
import collections
import difflib
import os
import pwd
import random
import subprocess
import sys
from fnmatch import fnmatch
import grp

original_mkdir = os.mkdir
original_listdir = os.listdir
original_open = __builtin__.open
original_check_output = subprocess.check_output
original_exists = os.path.exists
original_isdir = os.path.isdir
original_isfile = os.path.isfile
original_makedirs = os.makedirs
original_getsize = os.path.getsize
original_remove = os.remove
original_stat = os.stat
original_chmod = os.chmod
original_chown = os.chown
original_getpwnam = pwd.getpwnam
original_getgrnam = grp.getgrnam
_current_directory = os.path.dirname(os.path.realpath(__file__))


def read_data_file(name):
    filepath = os.path.join(_current_directory, 'datafiles', name)
    with open(filepath) as fi:
        return fi.read()


def write_data_file(name, target):
    with open(target, 'w') as fi:
        fi.write(read_data_file(name))


class FakeFile:

    def __init__(self, filepath, patcher, mode):
        self.filepath = filepath
        self.patcher = patcher
        self.mode = mode
        self.pos = 0
        if self.filepath not in self.patcher.file_expectations:
            self.patcher.file_expectations[self.filepath] = ''

    def read(self, length=-1):
        value = self.patcher.file_expectations[self.filepath]
        if length == -1:
            return value
        chunk = value[self.pos:self.pos + length]
        self.pos += len(chunk)
        return chunk

    def seek(self, pos):
        self.pos = pos

    def truncate(self, pos=-1):
        if pos == -1:
            pos = self.pos
        value = self.patcher.file_expectations[self.filepath]
        value = value[:pos]
        self.patcher.file_expectations[self.filepath] = value

    def __iter__(self):
        for line in self.read().splitlines():
            # weird thing that this splits on lines but keeps newline char
            yield line + '\n'

    def readlines(self):
        return self.read().splitlines()

    def write(self, data):
        if (self.filepath not in self.patcher.file_expectations or
                self.pos == 0):
            self.patcher.file_expectations[self.filepath] = ''
        self.patcher.file_expectations[self.filepath] += data
        self.pos += len(data)

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *ags):
        pass


PWType = collections.namedtuple('PW', ['pw_uid'])
Group = collections.namedtuple('Group', ['gr_gid'])
StatResult = collections.namedtuple(
    'Stat', ['st_mode', 'st_mtime', 'st_ino', 'st_size'])


class patched_system:

    def __init__(self):
        self.reset()

    def expect_command(self, command, output, returncode=0):
        if isinstance(command, list):
            command = ' '.join(command)
        self.cmd_expectations[command] = {
            'output': output,
            'returncode': returncode
        }

    def expect_file(self, filepath, value):
        self.file_expectations[filepath] = value

    def record(self, original, func=None):
        def new_func(*args, **kwargs):
            if original.__name__ not in self.records:
                self.records[original.__name__] = []
            self.records[original.__name__].append({
                'args': args,
                'kwargs': kwargs
            })
            if func is None:
                return
            else:
                return func(*args, **kwargs)

        return new_func

    def patch(self):
        from nclu import utils
        utils.subprocess.check_output = self.check_output
        __builtin__.open = self.open
        os.mkdir = self.record(os.mkdir)
        os.chmod = self.record(os.chmod)
        os.chown = self.record(os.chown)
        os.makedirs = lambda fn, mode: None
        os.listdir = self.listdir
        os.path.exists = self.exists
        os.path.isdir = self.isdir
        os.path.isfile = self.isfile
        os.path.getsize = self.getsize
        os.remove = self.remove
        os.stat = self.stat
        pwd.getpwnam = self.record(pwd.getpwnam, lambda u: PWType(1000))
        grp.getgrnam = self.record(grp.getgrnam, lambda g: Group(1000))

    def reset(self):
        self.cmd_expectations = collections.OrderedDict()
        self.file_expectations = {}
        self.command_runs = []
        self.records = {}
        self.expect_command(
            '/usr/lib/cumulus/cl-platform', 'cel,redstone_xp')
        self.expect_command(
            '/sbin/ifquery --syntax-help -o json',
            read_data_file('commands/ifquery/syntax-help'))

    def restore(self):
        from nclu import utils
        utils.subprocess.check_output = original_check_output
        os.mkdir = original_mkdir
        os.listdir = original_listdir
        __builtin__.open = original_open
        os.path.exists = original_exists
        os.path.isdir = original_isdir
        os.path.isfile = original_isfile
        os.makedirs = original_makedirs
        os.path.getsize = original_getsize
        os.remove = original_remove
        os.stat = original_stat
        os.chmod = original_chmod
        os.chown = original_chown
        pwd.getpwuid = original_getpwnam
        grp.getgrnam = original_getgrnam

    def remove(self, filepath):
        del self.file_expectations[filepath]

    def getsize(self, filepath):
        try:
            return len(self.file_expectations[filepath])
        except KeyError:
            return 0

    def exists(self, filepath):
        return self.isdir(filepath) or filepath in self.file_expectations

    def isdir(self, folder):
        folder = folder.rstrip('/')
        for filepath in self.file_expectations.keys():
            this_folder = '/'.join(filepath.split('/')[:-1])
            if folder != filepath and this_folder.startswith(folder):
                return True
        return False

    def isfile(self, filepath):
        return filepath in self.file_expectations

    def listdir(self, filepath):
        filepath = filepath.strip('/').split('/')
        values = []
        for other in self.file_expectations.keys():
            other = other.strip('/').split('/')
            if (other[:-1] == filepath and
                    len(other) - 1 == len(filepath)):
                values.append(other[-1])
        return values

    def open(self, filepath, mode='r', **kwargs):
        if filepath in self.file_expectations:
            return FakeFile(filepath, self, mode)
        if 'w' in mode:
            # we can write out to fake files...
            return FakeFile(filepath, self, mode)
        return original_open(filepath, mode, **kwargs)

    def stat(self, filepath):
        if filepath not in self.file_expectations:
            return original_stat(filepath)
        return StatResult(
            33152, 1534962189,
            str(random.randint(0, 99999)),
            len(self.file_expectations[filepath])
        )

    def run_diff(self, command, *args, **kwargs):
        '''
        override the diff getting run with system diff
        '''
        existing_filepath = command[-2]
        existing = self.file_expectations[command[-2]]
        new_filepath = command[-1]
        new = self.file_expectations[command[-1]]
        changes = []
        for part in difflib.unified_diff(
                    existing.splitlines(),
                    new.splitlines(),
                    fromfile=existing_filepath,
                    tofile=new_filepath,
                    n=10):
            changes.append(part)
        if len(changes) > 0:
            raise subprocess.CalledProcessError(
                1, command, '\n'.join(changes) + '\n')
        return ''

    def check_output(self, command, *args, **kwargs):
        '''
        patched check_output so we can control output
        '''
        fmt_command = command
        if isinstance(fmt_command, list):
            fmt_command = ' '.join(fmt_command)

        for reg_cmd, result in reversed(self.cmd_expectations.items()):
            if not fnmatch(fmt_command, reg_cmd):
                continue
            self.command_runs.append(command)
            if result['returncode'] == 0:
                return result['output']
            else:
                raise subprocess.CalledProcessError(
                    result['returncode'], result['output'])

        if fmt_command.startswith('/usr/bin/diff'):
            return self.run_diff(command, *args, **kwargs)

        raise Exception('Excepted {}'.format(fmt_command))


_installed_mocks = False


def install_system_mocks():
    global _installed_mocks
    if _installed_mocks:
        return
    directory = os.path.dirname(os.path.realpath(__file__))
    sys.path.append(os.path.join(directory, '_mocks'))
    _installed_mocks = True


# patcher should be global so it can be re-used between tests. Otherwise,
# modules that do the imports will be referencing the old patched method.
_patcher = None


def get_patcher():
    """
    Return the global instance of patched_system.  Initialize it first, if necessary.
    """

    global _patcher
    if _patcher is None:
        _patcher = patched_system()
        _patcher.patch()
    return _patcher
